function [gnew,hnew,bnew,sigma_one,p_one,q_one,CM]=...
    bulge_chasing(g,h,b,sigma,p,q,m,defl_size, alpha,n,CM)
% uses implicit QR wih compression to perform single shift bulge chasing on the
% submatrix m:defl_size;
% to be used in qrwc.m

% ****** Let us apply Theorem 3.1 ******

 %n=defl_size;
sigma_one=zeros(n-1,1);
Q=zeros(2,2,n-1); % Givens matrices will be stored here
hU=zeros(3,n); % uncompressed generators
gU=zeros(n,3); 
bU=zeros(3,3,n-1);
p_one=zeros(n,1); % updated vectors for rank-one correction
q_one=zeros(n,1);

% *** Parts 1 and 2: compute the Q's and update generators ***
% (new uncompressed generators are computed along with the Q's)

% ****** 1.1 ******

x=[g(1,:)*h(:,1)-p(1)*q(1)-alpha;sigma(1)-p(2)*q(1)];
G=planerot(x);  % this is Q_1'

   if m>1
       G=eye(2);
   end
   
Q(:,:,1)=G';
A=G*[g(1,:)*h(:,1), g(1,:)*b(:,:,1); sigma(1), g(2,:)];  % 2x3

B=[A(2,1), A(2,2:3)*h(:,2); p(3)*q(1), sigma(2)]*G';  % 2x2
pM=G*[p(1);p(2)];  % 2x1
qM=[q(1),q(2)]*G';  % 1x2
gU(1,:)=A(1,:);
hU(:,1)=[Q(1,1,1);h(:,2)*Q(2,1,1)];
bU(:,:,1)=[Q(1,2,1), 0, 0; h(:,2)*Q(2,2,1), b(:,:,2)];
p_one(1)=pM(1);
q_one(1)=qM(1);

% ****** 1.2 ******

for k=2:n-2
    x=[B(1,1)-pM(2)*qM(1); B(2,1)-p(k+1)*qM(1)];
   [G,Y]=planerot(x);
  
   if k>=defl_size
       G=eye(2);
       Y=x;
   end
  
   if k<m
       G=eye(2);
       Y=x;
   end
  
   if k==m
       xx=[g(k,:)*h(:,k)-p(k)*q(k)-alpha;sigma(k)-p(k+1)*q(k)];
       G=planerot(xx);
      
       Y=G*x;
   end

   Q(:,:,k)=G';
    A=G*[B(1,2), A(2,2:3)*b(:,:,k); B(2,2), g(k+1,:)];
    B=[A(2,1), A(2,2:3)*h(:,k+1); p(k+2)*qM(2), sigma(k+1)]*G';
    pM=G*[pM(2); p(k+1)];
    sigma_one(k-1)=Y(1)+pM(1)*qM(1);  % order is important!
    qM=[qM(2), q(k+1)]*G';
    gU(k,:)=A(1,:);
    hU(:,k)=[Q(1,1,k);h(:,k+1)*Q(2,1,k)];
    bU(:,:,k)=[Q(1,2,k), 0, 0; h(:,k+1)*Q(2,2,k), b(:,:,k+1)];
    p_one(k)=pM(1);
    q_one(k)=qM(1);
end

% ****** 1.3 ******

x=[B(1,1)-pM(2)*qM(1); B(2,1)-p(n)*qM(1)];
[G,Y]=planerot(x);

   if n-1>=defl_size
       G=eye(2);
       Y=x;
   end

   if (n-1)==m
       xx=[g(n-1,:)*h(:,n-1)-p(n-1)*q(n-1)-alpha;sigma(n-1)-p(n)*q(n-1)];
       G=planerot(xx);
       Y=G*x;
   end

Q(:,:,n-1)=G';
A=G*[B(1,2), A(2,2:3)*b(:,:,n-1); B(2,2), g(n,:)];
B=[A(2,1), A(2,2:3)*h(:,n)]*G';   % here only one row
pM=G*[pM(2); p(n)];
sigma_one(n-2)=Y(1)+pM(1)*qM(1);
qM=[qM(2), q(n)]*G';
sigma_one(n-1)=B(1);
p_one(n-1)=pM(1);
q_one(n-1)=qM(1);
gU(n-1,:)=A(1,:);
hU(:,n-1)=[Q(1,1,n-1);h(:,n)*Q(2,1,n-1)];
bU(:,:,n-1)=[Q(1,2,n-1),0,0; h(:,n)*Q(2,2,n-1),eye(2) ];  

% ****** 2 ******
% (last part of generator update)
p_one(n)=pM(2);
q_one(n)=qM(2);
gU(n,:)=[B(2),0,0];
hU(:,n)=[1;0;0];  


% *** CHECK: verify that the QR step has been performed correctly
%for k=1:n-1
%    CM(k:k+1,:)=Q(:,:,k)'*CM(k:k+1,:);
%    CM(:,k:k+1)=CM(:,k:k+1)*Q(:,:,k);
%end
%CMgen=compute_matrix(gU,hU,bU,sigma_one,p_one,q_one);
%norm(triu(CM)-triu(CMgen))


% ****** 3 ******
% (compression according to Theorem 2.1)
hnew=zeros(2,n);
gnew=zeros(n,2);
bnew=zeros(2,2,n-1);

Sold=hU(:,n);
hnew(1,n)=1;
gnew(n,1)=gU(n,:)*Sold;
hnew(:,n-1)=[1;0];
bnew(:,:,n-1)=[0,0;1,0];
Snew=[hU(:,n-1), bU(:,:,n-1)*Sold];
gnew(n-1,:)=gU(n-1,:)*Snew;
z=[sigma_one(n-1), gnew(n,1)];  
xold=p_one(n);  
for k=n-1:-1:2
   [V,Y]=planerot([p_one(k);xold]);  % here V = V_k'
   xnew=Y(1);
   A=V*[sigma_one(k-1), gnew(k,:); xold*q_one(k-1), z];
   xold=xnew;
   auxvec=A(2,:)';
  
   [G1,Y1]=planerot(auxvec(2:3));
   [G2,Y2]=planerot([auxvec(1);Y1(1)]);
   F=[G2,[0;0];0,0,1]*[1,0,0;[0;0],G1];  
   F=flipud(F);   % here F=F_k-1 (not *)
   hnew(:,k-1)=F(1:2,1);
   bnew(:,:,k-1)=F(1:2,2:3);
   Sold=Snew;
   Snew=hU(:,k-1)*(hnew(:,k-1))'+bU(:,:,k-1)*Sold*bnew(:,:,k-1)';
   z=A(1,1)*hnew(:,k-1)'+A(1,2:3)*bnew(:,:,k-1)';
   gnew(k-1,:)=gU(k-1,:)*Snew;
end
dummy=1;
%*** CHECK if compression works:
%CMcomp=compute_matrix(gnew,hnew,bnew,sigma_one,p_one,q_one);
%norm(CMgen-CMcomp)

